/*
** Network Audio Front-End (netaudio)
** Broadcasting routines
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <termios.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/ioctl.h>

#include "netaudio.h"

static int socks[16], sock;
char *header;
static fd_set readset, activeset;

static void broadcast_init(void)
{
    struct protoent *p;
    struct sockaddr_in sin;
    int i;

    for(i = 0; i < 16; i++)
	socks[i] = -1;

    bzero((char *) &sin, sizeof(sin));
    sin.sin_family = AF_INET;

    if (!strcmp(address, "127.0.0.1"))
	sin.sin_addr.s_addr = INADDR_ANY;
    else
	sin.sin_addr.s_addr = inet_addr(address);

    if ((sin.sin_port = htons(port)) == 0) {
	fprintf(stderr, "netaudio(broadcast): cannot translate port %d\n", port);
	exit(-1);
    }

    if ((p = getprotobyname("tcp")) == 0) {
	perror("getprotobyname(broadcast)");
	exit(-1);
    }

    if ((sock = socket(AF_INET, SOCK_STREAM, p->p_proto)) < 0) {
	perror("socket(broadcast)");
	exit(-1);
    }

    if (bind(sock, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
	perror("bind(broadcast)");
	exit(-1);
    }

    if (listen(sock, 5) < 0) {
	perror("listen(broadcast)");
	exit(-1);
    }

    fprintf(stderr, "netaudio: broadcast mode - waiting for connections.\n");
}

static int add_connection(int s)
{
    int i;

    for(i = 0; i < 16; i++)
	if (socks[i] == -1) {
	    socks[i] = s;
	    return(1);
	}

    return(0);
}

static void send_to_all(int size, char *buffer)
{
    int i;

    for(i = 0; i < 16; i++) {
	if (socks[i] != -1) {
	    if (write(socks[i], buffer, size) < 0) {
		if (!is_quiet)
		    fprintf(stderr, "netaudio: connection %d closed.\n", i);

		(void) shutdown(socks[i], 2);
		close(socks[i]);

		FD_CLR(socks[i], &activeset);
		FD_CLR(socks[i], &readset);

		socks[i] = -1;
	    }
	}
    }
}

static void handle_client(int i)
{
    int size, left = 0;
    char buffer[MAX_BUFFER], newbuffer[MAX_BUFFER];

    if (i == 0) {
reread:
	size = read(STDIN_FILENO, &newbuffer, frame_size - left);

	memcpy(buffer + left, newbuffer, size);
	left += size;

	if (size < 0) {
	    if (!is_quiet)
		fprintf(stderr, "netaudio(broadcast): connection closed.\n");

	    exit(-1);
	}

	if ((left < frame_size) && (size != 0))
	    goto reread;

	send_to_all(left, buffer);
    } else {
	i--;
	size = read(socks[i], &buffer, frame_size);

	if (size <= 0) {
	    if (!is_quiet)
		fprintf(stderr, "netaudio: connection %d closed.\n", i);

	    (void) shutdown(socks[i], 2);
	    close(socks[i]);

	    FD_CLR(socks[i], &readset);
	    FD_CLR(socks[i], &activeset);

	    socks[i] = -1;
	}
    }
}

void broadcast_loop(void)
{
    int fds = getdtablesize(), i;
    struct sockaddr_in sin;

    broadcast_init();

    FD_ZERO(&activeset);

    FD_SET(sock, &activeset);
    FD_SET(STDIN_FILENO, &activeset);

    while(1) {
	struct timeval t;

	t.tv_sec = 0;
	t.tv_usec = 100;

	bcopy(&activeset, &readset, sizeof(readset));

	if (select(fds, &readset, NULL, NULL, &t) < 0) {
	    if (!is_quiet)
		fprintf(stderr, "netaudio(broadcast): select error.\n");
	}

	if (FD_ISSET(sock, &readset)) {
	    int addrlen, newsock;

	    addrlen = sizeof(struct sockaddr_in);
	    newsock = accept(sock, (struct sockaddr *) &sin, &addrlen);

	    if (newsock < 0) {
		if (!is_quiet)
		    fprintf(stderr, "netaudio(broadcast): accept error.\n");
	    } else {
		if (!is_quiet)
		    fprintf(stderr, "netaudio(broadcast): incoming connection from %s\n", inet_ntoa(sin.sin_addr));

		FD_SET(newsock, &activeset);

		if (!add_connection(newsock)) {
		    (void) shutdown(newsock, 2);
		    close(newsock);

		    FD_CLR(newsock, &activeset);
		    FD_CLR(newsock, &readset);
		}
	    }
	}

	if (FD_ISSET(STDIN_FILENO, &readset))
	    handle_client(0);

	for(i = 0; i < 16; i++)
	    if (socks[i] != -1)
		if (FD_ISSET(socks[i], &readset))
		    handle_client(i + 1);
    }
}
